# -*- coding: UTF-8 -*-
"""
pysdd.__main__
~~~~~~~~~~~~~~

Command Line Interface.

Will be run when `python -m pysdd` is used.

:author: Wannes Meert, Arthur Choi
:copyright: Copyright 2017-2019 KU Leuven and Regents of the University of California.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
from .cli import main


if __name__ == "__main__":
    main()
